# SPDX-FileCopyrightText: (c) 2024 The Tenzir Contributors
# SPDX-License-Identifier: BSD-3-Clause

from pydantic import BaseModel
from typing import Literal, Optional


class RoleAndOrganizationRule(BaseModel):
    """
    Grants access to users who have some role in some organization.
    """

    auth_fn: Literal["auth_organization_role"] = "auth_organization_role"
    connection: Optional[str] = None

    # The required role.
    roles_claim: str
    role: str

    # The required organization.
    organization_claim: str
    organization: str


class OrganizationMembershipRule(BaseModel):
    """
    Grants access to users that belong to some organization.
    """

    auth_fn: Literal["auth_organization"] = "auth_organization"
    connection: Optional[str] = None
    organization_claim: str
    organization: str


class EmailDomainRule(BaseModel):
    """
    Grants access for users from a specific connection (e.g. google-oauth2)
    having email addresses from a given domain (e.g. @tenzir.com)
    """

    auth_fn: Literal["auth_email_suffix"] = "auth_email_suffix"
    connection: Optional[str] = None
    email_domain: str


class UserAuthRule(BaseModel):
    auth_fn: Literal["auth_user"] = "auth_user"
    user_id: str


AuthRule = (
    UserAuthRule
    | EmailDomainRule
    | RoleAndOrganizationRule
    | OrganizationMembershipRule
)



# SPDX-FileCopyrightText: (c) 2024 The Tenzir Contributors
# SPDX-License-Identifier: BSD-3-Clause

"""Usage:
  tenzir-platform admin add-auth-rule [-d] email-domain <workspace_id> <domain> [--connection=<connection>]
  tenzir-platform admin add-auth-rule [-d] organization-membership <workspace_id> <organization_claim> <organization> [--connection=<connection>]
  tenzir-platform admin add-auth-rule [-d] organization-role <workspace_id> <roles_claim> <role> <organization_claim> <organization> [--connection=<connection>]
  tenzir-platform admin add-auth-rule [-d] user <workspace_id> <user_id>
  tenzir-platform admin delete-auth-rule <workspace_id> <auth_rule_index>
  tenzir-platform admin list-auth-rules <workspace_id>
  tenzir-platform admin create-workspace <owner_namespace> <owner_id> [--name=<workspace_name>] [--category=<workspace_category>]
  tenzir-platform admin delete-workspace <workspace_id>
  tenzir-platform admin update-workspace <workspace_id> [--name=<workspace_name>] [--icon_url=<icon_url>] [--owner-namespace=<namespace>] [--owner-id=<owner_id>] [--category=<workspace_category>]
  tenzir-platform admin list-global-workspaces

Options:
  -d,--dry-run                      Do not add the rule, only print a JSON representation.
  --connection=<connection>         An optional prefix that must be matched by the 'sub' field
  --name=<workspace_name>           The user-visible name of the workspace.
  --icon_url=<icon_url>             The image to be used for this workspace in the frontend.
  --owner-id=<owner_id>             The owner id of the current namespace. This can be set to
                                    an arbitrary string that allows the platform administrators
                                    to identify the owner of this workspace.
  --owner-namespace=<namespace>     Must be either 'user' or 'organization'.
  --category=<workspace_category>   An arbitrary string that is used as header when grouping
                                    multiple workspaces from the same owner in the frontend.
                                    Note that currently only workspaces with the same owner id
                                    are grouped.
"""

from tenzir_platform.helpers.client import AppClient, TargetApi
from tenzir_platform.helpers.oidc import IdTokenClient
from tenzir_platform.helpers.environment import PlatformEnvironment
from tenzir_platform.helpers.auth_rule import (
    AuthRule,
    UserAuthRule,
    EmailDomainRule,
    RoleAndOrganizationRule,
    OrganizationMembershipRule,
)
from pydantic import BaseModel
from typing import Optional
from docopt import docopt


def _get_global_workspaces(client: AppClient):
    resp = client.get(
        "global-tenant-list",
        target_api=TargetApi.ADMIN,
    )
    resp.raise_for_status()
    return resp.json()


def add_auth_rule(client: AppClient, workspace_id: str, rule: AuthRule):
    resp = client.post(
        "add-auth-function",
        target_api=TargetApi.ADMIN,
        json={"tenant_id": workspace_id, "auth_fn": rule.model_dump()},
    )
    resp.raise_for_status()
    print(f"Added {rule.model_dump_json()}")


def delete_auth_rule(client: AppClient, workspace_id: str, index: int):
    resp = client.post(
        "delete-auth-function",
        target_api=TargetApi.ADMIN,
        json={"tenant_id": workspace_id, "index": index},
    )
    resp.raise_for_status()
    print(f"Deleted auth function {resp.json()} of {workspace_id}")


def list_auth_rules(client: AppClient, workspace_id: str):
    resp = client.get(
        "global-tenant-list",
        target_api=TargetApi.ADMIN,
    )
    resp.raise_for_status()
    for t in resp.json():
        if t["id"] == workspace_id:
            print(t["auth_functions"])


def create_workspace(client: AppClient, name: str, owner: str, owner_namespace: str):
    resp = client.post(
        "create-tenant",
        target_api=TargetApi.ADMIN,
        json={
            "owner": {
                "namespace": owner_namespace,
                "owner_id": owner,
            },
            "name": name,
        },
    )
    resp.raise_for_status()
    tenant_id = resp.json()["tenant_id"]
    print(f"Created workspace {tenant_id}")


def delete_workspace(client: AppClient, workspace_id: str):
    resp = client.post(
        "force-delete-tenant",
        target_api=TargetApi.ADMIN,
        json={
            "tenant_id": workspace_id,
        },
    )
    resp.raise_for_status()
    print(resp.json())


def update_workspace(
    client: AppClient,
    workspace_id: str,
    workspace_name: Optional[str],
    icon_url: Optional[str],
    owner_namespace: Optional[str],
    owner_id: Optional[str],
    workspace_category: Optional[str],
):
    resp = client.post(
        "update-tenant",
        target_api=TargetApi.ADMIN,
        json={
            "tenant_id": workspace_id,
            "name": workspace_name,
            "icon_url": icon_url,
            "owner_namespace": owner_namespace,
            "owner_id": owner_id,
            "display_name": workspace_category,
        },
    )
    resp.raise_for_status()
    print(f"updated workspace {workspace_id}")


def list_global_workspaces(client: AppClient):
    tenants = _get_global_workspaces(client)
    print(tenants)


def auth_rule_from_arguments(arguments) -> AuthRule:
    connection = arguments["--connection"]
    domain = arguments["<domain>"]
    user_id = arguments["<user_id>"]
    organization_claim = arguments["<organization_claim>"]
    organization = arguments["<organization>"]
    roles_claim = arguments["<roles_claim>"]
    role = arguments["<role>"]
    auth_rule: AuthRule
    if arguments["email-domain"]:
        auth_rule = EmailDomainRule(connection=connection, email_domain=domain)
    elif arguments["organization-membership"]:
        auth_rule = OrganizationMembershipRule(
            connection=connection,
            organization_claim=organization_claim,
            organization=organization,
        )
    elif arguments["organization-role"]:
        auth_rule = RoleAndOrganizationRule(
            connection=connection,
            roles_claim=roles_claim,
            role=role,
            organization_claim=organization_claim,
            organization=organization,
        )
    elif arguments["user"]:
        auth_rule = UserAuthRule(user_id=user_id)
    else:
        raise Exception("couldn't determine auth rule from command-line arguments")
    return auth_rule


def admin_subcommand(platform: PlatformEnvironment, argv):
    arguments = docopt(__doc__, argv=argv)

    def connect_and_login() -> AppClient:
        token_client = IdTokenClient(platform)
        token = token_client.load_id_token()
        client = AppClient(platform)
        client.user_login(token)
        return client

    if arguments["create-workspace"]:
        client = connect_and_login()
        owner_namespace = arguments["<owner_namespace>"]
        owner = arguments["<owner_id>"]
        name: str
        if arguments["--name"] is not None:
            name = arguments["--name"]
        else:
            name = f"{owner}'s Workspace"
        category = arguments["--category"]
        create_workspace(client, name, owner, owner_namespace)

    if arguments["delete-workspace"]:
        workspace_id = arguments["<workspace_id>"]
        client = connect_and_login()
        delete_workspace(client, workspace_id)

    if arguments["update-workspace"]:
        workspace_id = arguments["<workspace_id>"]
        client = connect_and_login()
        owner_namespace = arguments["--owner-namespace"]
        owner_id = arguments["--owner-id"]
        workspace_category = arguments["--category"]
        workspace_name = arguments["--name"]
        icon_url = arguments["--icon-url"]
        update_workspace(
            client,
            workspace_id,
            workspace_name=workspace_name,
            icon_url=icon_url,
            owner_namespace=owner_namespace,
            owner_id=owner_id,
            workspace_category=workspace_category,
        )

    if arguments["list-global-workspaces"]:
        client = connect_and_login()
        list_global_workspaces(client)

    if arguments["add-auth-rule"]:
        workspace_id = arguments["<workspace_id>"]
        dry_run = arguments["--dry-run"]
        auth_rule = auth_rule_from_arguments(arguments)
        if dry_run:
            print(f"Would add rule {auth_rule.model_dump_json()}")
        else:
            client = connect_and_login()
            add_auth_rule(client, workspace_id, auth_rule)

    if arguments["delete-auth-rule"]:
        workspace_id = arguments["<workspace_id>"]
        auth_fn_index: int = arguments["<auth_fn_index>"]
        client = connect_and_login()
        delete_auth_rule(client, workspace_id, auth_fn_index)

    if arguments["list-auth-rules"]:
        workspace_id = arguments["<workspace_id>"]
        client = connect_and_login()
        list_auth_rules(client, workspace_id)



# SPDX-FileCopyrightText: (c) 2024 The Tenzir Contributors
# SPDX-License-Identifier: BSD-3-Clause

"""
Usage:
  tenzir-platform workspace list [--json]
  tenzir-platform workspace select <workspace>
  tenzir-platform workspace rename <name>


Description:
  tenzir-platform workspace list
    Display a list of all workspaces accessible for the current user.

  tenzir-platform workspace select <workspace>
    Select <workspace>, which can be either a workspace id, a name or a number.

  tenzir-platform workspace rename <name>
    Rename the current workspace.

"""

from tenzir_platform.helpers.client import AppClient, TargetApi
from tenzir_platform.helpers.environment import PlatformEnvironment
from tenzir_platform.helpers.cache import store_workspace, load_current_workspace
from tenzir_platform.helpers.oidc import IdTokenClient
from docopt import docopt
from typing import List
import json
import re


def _is_workspace_id(identifier: str):
    return bool(re.match(r"^t-[a-z0-9]{8}$", identifier))


def _get_workspace_list(client: AppClient, id_token: str) -> List:
    resp = client.post(
        "get-login-info",
        json={
            "id_token": id_token,
        },
        target_api=TargetApi.USER_PUBLIC,
    )
    resp.raise_for_status()
    return resp.json()["allowed_tenants"]


def _resolve_workspace_identifier(
    client: AppClient, id_token: str, identifier: str
) -> str:
    # If we already have a node id, use that.
    if _is_workspace_id(identifier):
        return identifier

    # Otherwise go through the list of workspaces
    workspaces = _get_workspace_list(client, id_token)

    # If we're given a number, interpret it as a workspace index
    try:
        return workspaces[int(identifier)]["tenant_id"]
    except ValueError:
        pass
    except IndexError:
        raise Exception(f"Only have {len(workspaces)-1} workspaces")

    # Else, look for a matching workspace name
    name_matched = [
        workspace for workspace in workspaces if workspace["name"] == identifier
    ]
    if len(name_matched) == 0:
        raise Exception(f"Unknown workspace {identifier}")
    if len(name_matched) > 1:
        matching_ids = [workspace["tenant_id"] for workspace in name_matched]
        raise Exception(
            f"Ambigous name {identifier} is shared by workspaces {matching_ids}"
        )
    return name_matched[0]["tenant_id"]


def list(platform: PlatformEnvironment, print_json: bool):
    """Get list of authorized workspaces for the current CLI user"""
    id_token = IdTokenClient(platform).load_id_token()
    app_cli = AppClient(platform=platform)
    workspaces = _get_workspace_list(app_cli, id_token)
    if print_json:
        print(f"{json.dumps(workspaces)}")
    else:
        for i, workspace in enumerate(workspaces):
            print(f"{i}: {workspace['name']} ({workspace['tenant_id']}) ")


def select(platform: PlatformEnvironment, workspace_id_or_name: str):
    """Log in to a tenant as the current CLI user"""
    id_token = IdTokenClient(platform).load_id_token()
    app_cli = AppClient(platform)
    workspace_id = _resolve_workspace_identifier(
        app_cli, id_token, workspace_id_or_name
    )
    resp = app_cli.post(
        "switch-tenant",
        json={
            "id_token": id_token,
            "tenant_id": workspace_id,
        },
        target_api=TargetApi.USER_PUBLIC,
    )
    resp.raise_for_status()
    user_key = resp.json()["user_key"]
    store_workspace(platform, workspace_id, user_key)
    print(f"Switched to workspace {workspace_id}")


def rename(platform: PlatformEnvironment, workspace_id: str, name: str):
    app_cli = AppClient(platform)
    resp = app_cli.post(
        "rename-tenant",
        json={
            "tenant_id": workspace_id,
            "name": name,
        },
        target_api=TargetApi.USER_PUBLIC,
    )
    resp.raise_for_status()
    print(f"Renamed workspace {workspace_id}")


def workspace_subcommand(platform: PlatformEnvironment, argv):
    args = docopt(__doc__, argv=argv)
    if args["list"]:
        json = args["--json"]
        list(platform, json)
    if args["select"]:
        workspace_id = args["<workspace>"]
        select(platform, workspace_id)
    if args["rename"]:
        workspace_id, user_key = load_current_workspace(platform)
        name = args["<name>"]
        rename(platform=platform, workspace_id=workspace_id, name=name)



# SPDX-FileCopyrightText: (c) 2024 The Tenzir Contributors
# SPDX-License-Identifier: BSD-3-Clause

from pydantic_settings import BaseSettings
from typing import Optional

API_ENDPOINT = "https://rest.tenzir.app/production-v1"
ISSUER_URL = "https://tenzir.eu.auth0.com/"
CLIENT_ID = "vzRh8grIVu1bwutvZbbpBDCOvSzN8AXh"


class PlatformEnvironment(BaseSettings):
    # The remote API endpoint of the platform.
    api_endpoint: str = API_ENDPOINT

    # An arbitrary short string to identify this environment
    # in the local cache directory.
    stage_identifier: str = "prod"

    # TODO: Provide a new `/oidc-config` endpoint in the public platform api
    # to load these values dynamically.
    issuer_url: str = ISSUER_URL
    client_id: str = CLIENT_ID
    # Client secret is only necessary for non-interactive logins using the client
    # credentials flow.
    client_secret: Optional[str] = None
    client_secret_file: Optional[str] = None

    # Enable more verbose print statements.
    verbose: bool = False

    @staticmethod
    def load():
        return PlatformEnvironment(
            _env_prefix="TENZIR_PLATFORM_CLI_", _env_nested_delimiter="__"
        )



# SPDX-FileCopyrightText: (c) 2024 The Tenzir Contributors
# SPDX-License-Identifier: BSD-3-Clause

"""Usage:
  tenzir-platform auth login [--non-interactive]

Options:
  --non-interactive   Use client credentials flow for login.

Notes:
  When the `--non-interactive` the CLI will attempt to read a client
  secret from the `TENZIR_PLATFORM_CLI_CLIENT_SECRET[_FILE]` variable.
  This option is only useful for users running an on-prem instance of
  the platform who can configure a suitable client.
"""

from tenzir_platform.helpers.oidc import IdTokenClient
from tenzir_platform.helpers.environment import PlatformEnvironment
from docopt import docopt


def login(platform: PlatformEnvironment, interactive: bool):
    token_client = IdTokenClient(platform)
    token = token_client.load_id_token(interactive=interactive)
    decoded_token = token_client.validate_token(token)
    print(f"Logged in as {decoded_token.user_id}")


def auth_subcommand(platform: PlatformEnvironment, argv):
    args = docopt(__doc__, argv=argv)
    if args["login"]:
        interactive = not args["--non-interactive"]
        login(platform, interactive)
